"""
rich-codex is a minimal Python module for generating terminal screenshots from shell commands and text snippets.

It should work as a standalone command-line tool, however it is primarily intended
for use with the rich-codex GitHub Action.
"""

__version__ = "1.2.0"

from rich_codex import rich_img  # noqa: F401
